import os
from io import BytesIO
from pathlib import Path

import numpy as np
import pandas as pd
import requests
import typer
import validators

from eu_state_aids.utils import validate_year

app = typer.Typer()


@app.command()
def fetch(
    year: str,
    local_path: str = typer.Option(
        "./data/bg",
        help="Local path to use for eufunds excel files. Always use forward slashes."
    )
):
    """Fetch Excel file from the source and store it locally.

    Create the directory if it does not exist. Default directory is ./data/bg,
    and it can be changed with local_path.

    Forward slaches based paths are translated into proper paths using `pathlib`,
    so, even on Windows, there's no need to use backward slashes.
    """

    # script parameters validations
    assert(validate_year(year))

    # years are mapped to their codes (sic)
    years_encoding = {
        '2014': '3xRCSNcrgNc%3D',
        '2015': '8U%2BIPGXBzzM%3D',
        '2016': 'L35Wg8m16s0%3D',
        '2017': 'wxlx7atW%2FuQ%3D',
        '2018': 'DrjrB6YmlCo%3D',
        '2019': 'c3E0NC9D3EE%3D',
        '2020': 'ip23bQ8hOOQ%3D',
        '2021': 'mwei5Zc2UEA%3D',
        '2022': '2Q0LfOC9lQ4%3D',
        '2023': 'VuK356PVlaY%3D'
    }

    # create directory if not existing
    local_path = Path(local_path)
    if not os.path.exists(local_path):
        os.makedirs(local_path)

    # build remote url and filepath, out of the year,
    typer.echo(f"Fetching EU data for year: {year}")
    excel_url = f"http://2020.eufunds.bg/en/0/0/Project/ExportToExcel?StFrom={years_encoding[year]}&" \
        f"StTo={years_encoding[year]}&ShowRes=True&IsProgrammeSelected=False&IsRegionSelected=False"

    # read remote excel file content
    r = requests.get(excel_url)

    # store it locally
    filepath = local_path / f"projects_{year}.xlsx"
    with open(filepath, 'wb') as f:
        f.write(r.content)
    typer.echo(f"File saved to {filepath}")


@app.command()
def export(
    year: str,
    local_path: str = typer.Option(
        "./data/bg",
        help="Local path to use for eufunds excel files. Leave unset to use eufunds.org source."
    ),
    stateaid_url: str = typer.Option("https://stateaid.minfin.bg/document/860", help="URL of stateaid excel file"),
    program_start_year: str = typer.Option("2014", help="Program's starting year"),
):
    """Read Excel file from local path, produces CSV output in the same local path.

    Local path defaults to ./data/bg, and can be changed with local_path.
    Uses pandas to read from Excel, transform the dataframe, and cross the data with
    a second source to find out which of the EU funds are related to state aids.
    """

    # script parameters validations
    assert(validate_year(year))
    assert(validate_year(program_start_year))
    assert(validators.url(stateaid_url))

    # read the dataframe from the local file
    # the header starts at the 4th line
    # the pandas.DataFrame is created reading from the excel file's url
    typer.echo(f"Fetching EU data for year: {year}")
    local_path = Path(local_path)
    excel_file = local_path / f"projects_{year}.xlsx"
    eu_df = pd.read_excel(f"file://localhost/{os.path.abspath(excel_file)}", header=3)

    # The last 6 rows are removed from the dataframe, as they contain the notes
    eu_df = eu_df[:-6]
    typer.echo(f"DataFrame with {len(eu_df)} rows created from {excel_file}")

    # EU dataframe transformations

    def lookup_name(x):
        try:
            return x.split(' ', 1)[1]
        except IndexError:
            return x
    eu_df['Name of the beneficiary'] = eu_df.Beneficiary.apply(lookup_name)

    def lookup_id(x):
        try:
            return x.split(' ', 1)[0]
        except IndexError:
            return np.NAN
    eu_df['ID of the beneficiary'] = eu_df.Beneficiary.apply(lookup_id)

    def build_eu_prog_id(x):
        try:
            return program_start_year + x.split('-', 1)[0]
        except AttributeError:
            return np.NAN
    eu_df['European operation program (ID)'] = eu_df['Project proposal number'].apply(build_eu_prog_id)

    eu_df['Amounts (€)'] = eu_df['Total'] * 0.51
    eu_df['Amounts (€)'] = eu_df['Amounts (€)'].astype(float).round(decimals=2)

    eu_df['Date'] = year

    # the stateaid file is fetched (download)
    typer.echo(f"Fetching stateaid data at {stateaid_url}")
    r = requests.get(stateaid_url)

    # DataFrame is created out of a stream, generated by a content
    # the header starts at the 1st line
    # only the first 10 columns are kept
    stateaid_df = pd.read_excel(BytesIO(r.content), header=1)
    stateaid_df = stateaid_df.iloc[:, :10]

    def build_eu_prog_aid_id(x):
        try:
            return '-'.join(x.split('-')[:-1])
        except AttributeError:
            return np.NAN
    eu_df.dropna(subset=['Project proposal number'], inplace=True)
    eu_df['EUProgAidID'] = eu_df['Project proposal number'].apply(build_eu_prog_aid_id)
    eu_df.dropna(subset=['EUProgAidID'], inplace=True)

    def lookup_prog_id(x):
        """Looks up for program ID in stateaid_df

        :param x:
        :return:
        """
        has_x = stateaid_df[stateaid_df.iloc[:, 1].str.contains(x)]
        if not has_x.empty:
            return has_x.iloc[0, 0].split(',')[0]
        else:
            return np.NAN
    eu_df['State aid Scheme'] = eu_df['EUProgAidID'].apply(lookup_prog_id)
    eu_df.dropna(subset=['State aid Scheme'], inplace=True)

    eu_df.drop(
        columns=[
            'Beneficiary', 'Address', 'Location', 'Project proposal number',
            'Project Name', 'Total', 'Grant', 'Self-financing by the Beneficiary',
            'Actual amounts paid', 'Duration (months)',
            'Status of Implementation of the Contract/Order of the Grant',
            'EUProgAidID'
        ],
        inplace=True
    )

    # emit csv
    typer.echo(f"{len(eu_df)} matches found.")
    csv_filepath = local_path / f"{year}.csv"
    if len(eu_df):
        typer.echo(f"Writing results to {csv_filepath}")
        eu_df.to_csv(csv_filepath, na_rep='', index=False)



import validators
from validators.utils import validator


@validator
def validate_year(y: str) -> bool:
    """Validate y as a valid year

    :param y: the year as a string
    :return: if the year is valid or not
    """
    try:
        y = int(y)
    except ValueError:
        return False

    return validators.between(y, min=2010, max=2050)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eu_state_aids']

package_data = \
{'': ['*']}

install_requires = \
['importlib-metadata>=4.5.0,<5.0.0',
 'openpyxl>=3.0.7,<4.0.0',
 'pandas>=1.2.4,<2.0.0',
 'requests-mock>=1.9.3,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'typer>=0.3.2,<0.4.0',
 'validators>=0.18.2,<0.19.0']

entry_points = \
{'console_scripts': ['eu-state-aids = eu_state_aids.main:app']}

setup_kwargs = {
    'name': 'eu-state-aids',
    'version': '0.2.0',
    'description': 'CLI to extract state aids data from public sources and produce CSV files',
    'long_description': '## Description\n\n`eu-state-aids` is a Command Line Interface that can be used \nto import **state aids related data** from single countries sources sitesm\nand produce CSV files, according to a common  data structure.\n\n\n![TravisCI Badge](https://travis-ci.com/openpolis/eu-state-aids.svg?branch=master "TravisCI building status")\n[![PyPI version](https://badge.fury.io/py/eu-state-aids.svg)](https://badge.fury.io/py/eu-state-aids)\n![Tests Badge](https://op-badges.s3.eu-west-1.amazonaws.com/eu-state-aids/tests-badge.svg)\n![Coverage Badge](https://op-badges.s3.eu-west-1.amazonaws.com/eu-state-aids/coverage-badge.svg)\n\n## Installation\n\nPython versions from 3.7 are supported.\n \nThe package depends on these python packages:\n* typer\n* openpyxl\n* pandas\n* requests\n* validators\n\nSo, it\'s better to create a *virtualenv* before installation.\n\nThe package is hosted on pypi, and can be installed, for example using pip:\n\n    pip install eu-state-aids \n\n\n## Usage\n\nThe `eu-state-aids` binary command will be available after installation. \nIt offers help with:\n\n    eu-state-aids --help\n\nThe `opstate-aids` command can be used to extract the data from the official sources, \nand populate the CSV files.\n\nFor each country, data files will firstly be *fetched* and stored locally, \nand thereafter *used* in order to **export** CSV files.\n\nThis two-step procedure is useful, since it is not always possible to download source files (Excel, XML, ...) from \nBI systems of nation states, as it has been seen that they tend to time-out whenever the number of records is \nhigh enough.\n\nThe logic of these two phases can vary for each single european state, so each country will have a dedicated module,\nthat will be executable as a sub-command.\n\nTo retrieve data and produce a CSV file for Bulgary (bg), 2015:\n \n      eu-state-aids bg fetch 2015\n      eu-state-aids bg export 2015\n\nTo launch the scripts *for all years* for Bulgary (bg):\n\n    # download all years\' excel files into local storage \n    for Y in $(seq 2014 2022)\n    do \n      eu-state-aids bg fetch $Y\n    done\n    \n    # process all years\' excel files and export CSV records into local storage \n    #./data/bg/$Y.csv files\n    for Y in $(seq 2014 2022)\n      eu-state-aids bg export $Y\n    done\n\n## Support\n\nThere is no guaranteed support available, but issues can be created on this project \nand the authors will try to answer and merge proposed solutions into the code base.\n\n## Project Status\nThis project is funded by the European Commission and is currently (2021) under active developement.\n\n## Contributing\nIn order to contribute to this project:\n* verify that python 3.7+ is being used (or use [pyenv](https://github.com/pyenv/pyenv))\n* verify or install [poetry](https://python-poetry.org/), to handle packages and dependencies in a leaner way, \n  with respect to pip and requirements\n* clone the project `git clone git@github.com:openpolis/eu-state-aids.git` \n* install the dependencies in the virtualenv, with `poetry install`,\n  this will also install the dev dependencies\n* develop wildly, running tests and coverage with `coverage run -m pytest`\n\n### Testing\nTests are under the tests folder. [requests-mock](https://requests-mock.readthedocs.io/en/latest/index.html)\nis used to mock requests to remote data files, in order to avoid slow remote connections during tests.\n\n## Authors\nGuglielmo Celata - guglielmo@openpolis.it\n\n## Licensing\nThis package is released under an MIT License, see details in the LICENSE.txt file.\n\n',
    'author': 'guglielmo',
    'author_email': 'guglielmo@openpolis.it',
    'maintainer': 'guglielmo',
    'maintainer_email': 'guglielmo@openpolis.it',
    'url': 'https://github.com/openpolis/eu-state-aids/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)
